/*
 *   cdda - CD Digital Audio support
 *
 *   Copyright (C) 1993-2002  Ti Kan
 *   E-mail: xmcd@amb.org
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *   SGI Irix sound driver support
 */
#ifndef LINT
static char *_wr_irix_c_ident_ = "@(#)wr_irix.c	7.17 02/11/13";
#endif

#include "common_d/appenv.h"
#include "common_d/util.h"
#include "libdi_d/libdi.h"
#include "cdda_d/cdda.h"

#if defined(CDDA_WR_IRIX) && defined(CDDA_SYSVIPC)

#include <dmedia/audio.h>
#include "cdda_d/sysvipc.h"
#include "cdda_d/sem.h"
#include "cdda_d/wr_gen.h"
#include "cdda_d/wr_irix.h"


#define OUT_RESNAME	".AnalogOut"

extern appdata_t	app_data;
extern FILE		*errfp;

STATIC ALconfig		irix_config;		/* Audio config */
STATIC ALport		irix_port = NULL;	/* Audio port */
STATIC int		irix_wsemid;		/* Semaphores identifier */
STATIC pid_t		irix_pipe_pid = -1;	/* Pipe to program pid */
STATIC cd_state_t	*irix_wcd;		/* Shared memory pointer */
STATIC bool_t		irix_write_dev,		/* Write to audio device */
			irix_write_file,	/* Write to output file */
			irix_write_pipe,	/* Pipe to program */
			irix_file_be = FALSE;	/* Big endian output file */
STATIC gen_desc_t	*irix_file_desc = NULL,	/* Output file desc */
			*irix_pipe_desc = NULL;	/* Pipe to program desc */
STATIC unsigned int	irix_file_mode;		/* Output file mode */


/*
 * irixaud_open_dev
 *	Opens the audio device.
 *
 * Args:
 *	None.
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
STATIC bool_t
irixaud_open_dev(void)
{
	int	audiodev;
	char	*cp;

	if (!gen_set_eid(irix_wcd))
		return FALSE;

	if ((irix_config = alNewConfig()) == 0) {
		(void) sprintf(irix_wcd->i->msgbuf,
			    "irixaud_open_dev: audio config failed:\n%s",
			    alGetErrorString(oserror()));
		DBGPRN(DBG_SND)(errfp, "%s\n", irix_wcd->i->msgbuf);
		(void) gen_reset_eid(irix_wcd);
		return FALSE;
	}

	/* Set audio parameters */
	(void) alSetSampFmt(irix_config, AL_SAMPFMT_TWOSCOMP);
	(void) alSetWidth(irix_config, AL_SAMPLE_16);
	(void) alSetChannels(irix_config, AL_STEREO);
	(void) alSetQueueSize(irix_config, 131069);

	if ((cp = getenv("AUDIODEV")) == NULL) {
		/* Use default output port */
		audiodev = AL_DEFAULT_OUTPUT;
	}
	else {
		char	*devname;

		devname = (char *) MEM_ALLOC(
			"devname",
			strlen(cp) + strlen(OUT_RESNAME) + 1
		);
		if (devname == NULL) {
			(void) sprintf(irix_wcd->i->msgbuf,
				    "irixaud_open_dev: Out of memory");
			DBGPRN(DBG_SND)(errfp, "%s\n", irix_wcd->i->msgbuf);
			(void) gen_reset_eid(irix_wcd);
			return FALSE;
		}

		(void) sprintf(devname, "%s%s", cp, OUT_RESNAME);

		if ((audiodev = alGetResourceByName(AL_SYSTEM, devname,
						    AL_DEVICE_TYPE)) == 0) {
			(void) sprintf(irix_wcd->i->msgbuf,
				    "irixaud_open_dev: "
				    "Invalid audio resource: %s\n%s",
				    devname, alGetErrorString(oserror()));
			MEM_FREE(devname);
			DBGPRN(DBG_SND)(errfp, "%s\n", irix_wcd->i->msgbuf);
			(void) gen_reset_eid(irix_wcd);
			return FALSE;
		}

		MEM_FREE(devname);
	}

	/* Set the device */
	if (alSetDevice(irix_config, audiodev) < 0) {
		(void) sprintf(irix_wcd->i->msgbuf,
			    "irixaud_open_dev: Cannot set device\n%s",
			    alGetErrorString(oserror()));
		DBGPRN(DBG_SND)(errfp, "%s\n", irix_wcd->i->msgbuf);
		(void) gen_reset_eid(irix_wcd);
		return FALSE;
	}

	/* Open the port */
	if ((irix_port = alOpenPort("xmcd-writer", "w", irix_config)) == NULL){
		(void) sprintf(irix_wcd->i->msgbuf,
			    "irixaud_open_dev: Cannot open port\n%s",
			    alGetErrorString(oserror()));
		DBGPRN(DBG_SND)(errfp, "%s\n", irix_wcd->i->msgbuf);
		(void) gen_reset_eid(irix_wcd);
		return FALSE;
	}

	DBGPRN(DBG_GEN|DBG_SND)(errfp, "\nOpened audio device [%d]\n",
				audiodev);

	(void) gen_reset_eid(irix_wcd);
	return TRUE;
}


/*
 * irixaud_close_dev
 *	Closes the audio device.
 *
 * Args:
 *	None.
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
STATIC bool_t
irixaud_close_dev(void)
{
	if (!gen_set_eid(irix_wcd))
		return FALSE;

	if (irix_port != NULL) {
		/* Drain audio */
		while (alGetFilled(irix_port) > 0)
			util_delayms(10);

		alClosePort(irix_port);
		alFreeConfig(irix_config);
		irix_port = NULL;

		DBGPRN(DBG_GEN|DBG_SND)(errfp, "\nClosed audio device\n");
	}

	(void) gen_reset_eid(irix_wcd);
	return TRUE;
}


/*
 * irixaud_config
 *	Sets up audio device for playing CD quality audio.
 *
 * Args:
 *	None.
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
STATIC bool_t
irixaud_config(void)
{
	int	audiodev;
	ALpv	p[1];

	if ((audiodev = alGetDevice(irix_config)) < 0) {
		(void) sprintf(irix_wcd->i->msgbuf,
			    "irixaud_config: Cannot get device:\n%s",
			    alGetErrorString(oserror()));
		DBGPRN(DBG_SND)(errfp, "%s\n", irix_wcd->i->msgbuf);
		return FALSE;
	}

	if (alSetSampFmt(irix_config, AL_SAMPFMT_TWOSCOMP) < 0) {
		(void) sprintf(irix_wcd->i->msgbuf,
			    "irixaud_config: Cannot set sampling format:\n%s",
			    alGetErrorString(oserror()));
		DBGPRN(DBG_SND)(errfp, "%s\n", irix_wcd->i->msgbuf);
		return FALSE;
	}

	if (alSetWidth(irix_config, AL_SAMPLE_16) < 0) {
		(void) sprintf(irix_wcd->i->msgbuf,
			    "irixaud_config: Cannot set width:\n%s",
			    alGetErrorString(oserror()));
		DBGPRN(DBG_SND)(errfp, "%s\n", irix_wcd->i->msgbuf);
		return FALSE;
	}

	if (alSetChannels(irix_config, AL_STEREO) < 0) {
		(void) sprintf(irix_wcd->i->msgbuf,
			    "irixaud_config: Cannot set channels:\n%s",
			    alGetErrorString(oserror()));
		DBGPRN(DBG_SND)(errfp, "%s\n", irix_wcd->i->msgbuf);
		return FALSE;
	}

	p[0].param = AL_OUTPUT_RATE;
	p[0].value.ll = alDoubleToFixed(44100.0);

	if (alSetParams(audiodev, p, 1) < 0) {
		(void) sprintf(irix_wcd->i->msgbuf,
			    "irixaud_config: Cannot set sampling rate:\n%s",
			    alGetErrorString(oserror()));
		DBGPRN(DBG_SND)(errfp, "%s\n", irix_wcd->i->msgbuf);
		return FALSE;
	}

	return TRUE;
}


/*
 * irixaud_write
 *	Function responsible for writing from the buffer to the audio
 *	device (and file if saving).
 *
 * Args:
 *	s - Pointer to the curstat_t structure
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
STATIC bool_t
irixaud_write(curstat_t *s)
{
	byte_t	*in_data,
		*sw_data,
		*le_data,
		*be_data,
		*dev_data,
		*file_data,
		*pipe_data;
	char	*fname = NULL;
	int	trk_idx = -1,
		hbcnt = 1;
	time_t	start_time,
		tot_time;
	bool_t	ret;

	/* Get memory for audio device write buffer */
	le_data = (byte_t *) MEM_ALLOC("le_data", irix_wcd->cds->chunk_bytes);
	if (le_data == NULL) {
		(void) sprintf(irix_wcd->i->msgbuf,
				"irixaud_write: out of memory");
		DBGPRN(DBG_GEN)(errfp, "%s\n", irix_wcd->i->msgbuf);
		return FALSE;
	}

	/* Get memory for audio file write buffer */
	be_data = (byte_t *) MEM_ALLOC("be_data", irix_wcd->cds->chunk_bytes);
	if (be_data == NULL) {
		(void) sprintf(irix_wcd->i->msgbuf,
				"irixaud_write: out of memory");
		DBGPRN(DBG_GEN)(errfp, "%s\n", irix_wcd->i->msgbuf);
		return FALSE;
	}

	(void) memset(le_data, 0, irix_wcd->cds->chunk_bytes);
	(void) memset(be_data, 0, irix_wcd->cds->chunk_bytes);

	in_data = app_data.cdda_bigendian ? be_data : le_data;
	sw_data = app_data.cdda_bigendian ? le_data : be_data;

	/* Start time */
	start_time = time(NULL);

	/* While not stopped */
	ret = TRUE;
	while (irix_wcd->i->state != CDSTAT_COMPLETED) {
		/* Set the current track and track length info */
		cdda_wr_setcurtrk(irix_wcd);

		/* Get lock */
		cdda_waitsem(irix_wsemid, LOCK);

		/* End if finished and nothing in the buffer */
		if (irix_wcd->cdb->occupied == 0 && irix_wcd->i->cdda_done) {
			irix_wcd->i->state = CDSTAT_COMPLETED;
			cdda_postsem(irix_wsemid, LOCK);
			break;
		}

		/* Wait until there is something there */
		while (irix_wcd->cdb->occupied <= 0 &&
		       irix_wcd->i->state != CDSTAT_COMPLETED) {
			cdda_postsem(irix_wsemid, LOCK);
			cdda_waitsem(irix_wsemid, DATA);
			cdda_waitsem(irix_wsemid, LOCK);
		}

		/* Break if stopped */
		if (irix_wcd->i->state == CDSTAT_COMPLETED) {
			cdda_postsem(irix_wsemid, LOCK);
			break;
		}

		/* Copy a chunk from the nextout slot into the data buffers */
		(void) memcpy(
			in_data,
			&irix_wcd->cdb->b[
			    irix_wcd->cds->chunk_bytes * irix_wcd->cdb->nextout
			],
			irix_wcd->cds->chunk_bytes
		);

		/* Update pointers */
		irix_wcd->cdb->nextout++;
		irix_wcd->cdb->nextout %= irix_wcd->cds->buffer_chunks;
		irix_wcd->cdb->occupied--;

		/* Set heartbeat and interval, update stats */
		if (--hbcnt == 0) {
			cdda_heartbeat(&irix_wcd->i->writer_hb,CDDA_HB_WRITER);

			tot_time = time(NULL) - start_time;
			if (tot_time > 0 && irix_wcd->i->frm_played > 0) {
				irix_wcd->i->frm_per_sec = (int)
					((float) irix_wcd->i->frm_played /
					 tot_time) + 0.5;
			}

			hbcnt = cdda_heartbeat_interval(
				irix_wcd->i->frm_per_sec
			);
		}

		/* Signal room available */
		cdda_postsem(irix_wsemid, ROOM);

		/* Release lock */
		cdda_postsem(irix_wsemid, LOCK);

		/* Change channel routing if necessary */
		gen_chroute(irix_wcd->i->chroute, (sword16_t *)(void *) in_data,
			    (size_t) irix_wcd->cds->chunk_bytes);

		/* Prepare byte-swapped version of the data */
		gen_byteswap(in_data, sw_data,
			     (size_t) irix_wcd->cds->chunk_bytes);

		/* Write to device */
		if (irix_write_dev) {
			/* Always send big endian data */
			dev_data = be_data;

			if (alWriteFrames(irix_port, dev_data,
					irix_wcd->cds->chunk_bytes >> 2) < 0) {
				irix_wcd->i->state = CDSTAT_COMPLETED;
				cdda_postsem(irix_wsemid, LOCK);
				ret = FALSE;
				break;
			}
		}

		/* Write to file */
		if (irix_write_file) {
			file_data = irix_file_be ? be_data : le_data;

			/* Open new track file if necessary */
			if (!app_data.cdda_trkfile) {
				fname = s->trkinfo[0].outfile;
			}
			else if (trk_idx != irix_wcd->i->trk_idx) {
				trk_idx = irix_wcd->i->trk_idx;

				if (irix_file_desc != NULL &&
				    !gen_close_file(irix_file_desc)) {
					irix_wcd->i->state = CDSTAT_COMPLETED;
					cdda_postsem(irix_wsemid, LOCK);
					ret = FALSE;
					break;
				}

				fname = s->trkinfo[trk_idx].outfile;
				irix_file_desc = gen_open_file(
					irix_wcd,
					fname,
					O_WRONLY | O_TRUNC | O_CREAT,
					(mode_t) irix_file_mode,
					app_data.cdda_filefmt,
					irix_wcd->i->trk_len
				);
				if (irix_file_desc == NULL) {
					irix_wcd->i->state = CDSTAT_COMPLETED;
					cdda_postsem(irix_wsemid, LOCK);
					ret = FALSE;
					break;
				}
			}

			if (!gen_write_chunk(irix_file_desc, file_data,
					(size_t) irix_wcd->cds->chunk_bytes)) {
				irix_wcd->i->state = CDSTAT_COMPLETED;
				cdda_postsem(irix_wsemid, LOCK);
				ret = FALSE;
				break;
			}
		}

		/* Pipe to program */
		if (irix_write_pipe) {
			pipe_data = irix_file_be ? be_data : le_data;

			if (!gen_write_chunk(irix_pipe_desc, pipe_data,
					(size_t) irix_wcd->cds->chunk_bytes)) {
				irix_wcd->i->state = CDSTAT_COMPLETED;
				cdda_postsem(irix_wsemid, LOCK);
				ret = FALSE;
				break;
			}
		}

		irix_wcd->i->frm_played += irix_wcd->cds->chunk_blocks;

		while (irix_wcd->i->state == CDSTAT_PAUSED)
			util_delayms(400);

		/* Update debug level */
		app_data.debug = irix_wcd->i->debug;
	}

	/* Wake up reader */
	cdda_postsem(irix_wsemid, ROOM);

	/* Free memory */
	MEM_FREE(le_data);
	MEM_FREE(be_data);

	return (ret);
}


/*
 * irixaud_cleanup
 *	Detaches shared memory.
 *
 * Args:
 *	killreader - Whether to kill the reader process
 *
 * Return:
 *	Nothing.
 */
STATIC void
irixaud_cleanup(bool_t killreader)
{
	DBGPRN(DBG_GEN)(errfp,
			  "\nirixaud_cleanup: Cleaning up writer pid=%d\n",
			  (int) getpid());

	if (irix_port != NULL)
		(void) irixaud_close_dev();

	if (irix_file_desc != NULL) {
		(void) gen_close_file(irix_file_desc);
		irix_file_desc = NULL;
	}

	if (irix_pipe_desc != NULL) {
		(void) gen_close_pipe(irix_pipe_desc);
		irix_pipe_desc = NULL;
	}

	(void) sleep(1);

	if (irix_pipe_pid > 0) {
		(void) kill(irix_pipe_pid, SIGTERM);
		irix_pipe_pid = -1;
	}

	if (irix_wcd != NULL) {
		if (irix_wcd->i != NULL) {
			if (killreader && irix_wcd->i->reader != 0) {
				(void) kill(irix_wcd->i->reader, SIGTERM);
				irix_wcd->i->state = CDSTAT_COMPLETED;
			}

			/* Reset frames played counter */
			irix_wcd->i->frm_played = 0;
		}

		MEM_FREE(irix_wcd);
		irix_wcd = NULL;
	}
}


/*
 * irixaud_onterm
 *	SIGTERM and SIGPIPE signal handler - gracefully exit
 *
 * Args:
 *	signo - The signal number
 *
 * Return:
 *	Nothing.
 */
STATIC void
irixaud_onterm(int signo)
{
	irixaud_cleanup((bool_t) (signo == SIGPIPE));
	exit(1);
}


/*
 * irix_writeinit
 *	Pre-playback support check function
 *
 * Args:
 *	None.
 *
 * Return:
 *	Bitmask of supported features
 */
word32_t
irix_writeinit(void)
{
	return (CDDA_WRITEDEV | CDDA_WRITEFILE | CDDA_WRITEPIPE);
}


/*
 * irix_write
 *	Opens audio device/file, attaches shared memory and semaphores.
 *	Continuously reads data from shared memory and writes it
 *	to the audio device/file.
 *
 * Args:
 *	s    - Pointer to the curstat_t structure
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
bool_t
irix_write(curstat_t *s)
{
	size_t	estlen;

	if (!PLAYMODE_IS_CDDA(app_data.play_mode)) {
		(void) fprintf(errfp, "irix_write: Nothing to do.\n");
		return FALSE;
	}

	irix_write_dev  = (bool_t) ((app_data.play_mode & PLAYMODE_CDDA) != 0);
	irix_write_file = (bool_t) ((app_data.play_mode & PLAYMODE_FILE) != 0);
	irix_write_pipe = (bool_t) ((app_data.play_mode & PLAYMODE_PIPE) != 0);

	/* Generic services initialization */
	gen_write_init();

	/* Set some signal behaviors */
	(void) signal(SIGINT, SIG_IGN);
	(void) signal(SIGQUIT, SIG_IGN);
	(void) signal(SIGALRM, SIG_IGN);
	(void) signal(SIGUSR1, SIG_IGN);
	(void) signal(SIGPIPE, irixaud_onterm);
	(void) signal(SIGTERM, irixaud_onterm);

	/* Allocate memory */
	irix_wcd = (cd_state_t *) MEM_ALLOC("cd_state_t", sizeof(cd_state_t));
	if (irix_wcd == NULL) {
		(void) fprintf(errfp, "irix_write: out of memory\n");
		return FALSE;
	}

	(void) memset(irix_wcd, 0, sizeof(cd_state_t));

	/* Initialize shared memory and semaphores */
	if ((irix_wsemid = cdda_rw_initipc(irix_wcd)) < 0) {
		irixaud_cleanup(TRUE);
		return FALSE;
	}

	(void) sscanf(app_data.cdinfo_filemode, "%o", &irix_file_mode);
	/* Make sure file is at least accessible by user */
	irix_file_mode |= S_IRUSR | S_IWUSR;
	irix_file_mode &= ~(S_ISUID | S_ISGID | S_IWGRP | S_IWOTH);

	estlen = (irix_wcd->i->end_lba - irix_wcd->i->start_lba + 1) *
		 CDDA_BLKSZ;

	/* Set heartbeat */
	cdda_heartbeat(&irix_wcd->i->writer_hb, CDDA_HB_WRITER);

	if (irix_write_file || irix_write_pipe) {
		/* Open file and/or pipe */
		if (!app_data.cdda_trkfile && irix_write_file) {
			irix_file_desc = gen_open_file(
				irix_wcd,
				s->trkinfo[0].outfile,
				O_WRONLY | O_TRUNC | O_CREAT,
				(mode_t) irix_file_mode,
				app_data.cdda_filefmt,
				estlen
			);
			if (irix_file_desc == NULL) {
				irixaud_cleanup(TRUE);
				return FALSE;
			}
		}

		if (irix_write_pipe) {
			irix_pipe_desc = gen_open_pipe(
				irix_wcd,
				app_data.pipeprog,
				&irix_pipe_pid,
				app_data.cdda_filefmt,
				estlen
			);
			if (irix_pipe_desc == NULL) {
				irixaud_cleanup(TRUE);
				return FALSE;
			}
			DBGPRN(DBG_SND)(errfp,
					  "\nirixaud_write: Pipe to [%s]: "
					  "chunk_blks=%d\n",
					  app_data.pipeprog,
					  irix_wcd->cds->chunk_blocks);
		}

		switch (app_data.cdda_filefmt) {
		case FILEFMT_AU:
		case FILEFMT_AIFF:
		case FILEFMT_AIFC:
			/* Always big endian */
			irix_file_be = TRUE;
			break;
		case FILEFMT_WAV:
		case FILEFMT_OGG:
			/* Always little endian */
			irix_file_be = FALSE;
			break;
		case FILEFMT_MP3:
#ifdef USE_LIBMP3LAME
			/* Use native endian */
#if _BYTE_ORDER_ == _B_ENDIAN_
			irix_file_be = TRUE;
#else
			irix_file_be = FALSE;
#endif
#else
			/* Use big endian */
			irix_file_be = TRUE;
#endif
			break;
		default:
			break;
		}
	}

	if (irix_write_dev) {
		/* Open audio device */
		if (!irixaud_open_dev()) {
			irixaud_cleanup(TRUE);
			return FALSE;
		}

		/* Configure audio */
		if (!irixaud_config()) {
			irixaud_cleanup(TRUE);
			return FALSE;
		}
	}

	/* Do the writing */
	if (!irixaud_write(s)) {
		irixaud_cleanup(TRUE);
		return FALSE;
	}

	if (irix_write_file && !gen_close_file(irix_file_desc)) {
		irixaud_cleanup(TRUE);
		return FALSE;
	}
	irix_file_desc = NULL;

	if (irix_write_pipe && !gen_close_pipe(irix_pipe_desc)) {
		irixaud_cleanup(TRUE);
		return FALSE;
	}
	irix_pipe_desc = NULL;
	irix_pipe_pid = -1;

	irixaud_cleanup(FALSE);
	return TRUE;
}


/*
 * irix_winfo
 *	Append method identification to informational string.
 *
 * Args:
 *	str - The informational string
 *
 * Return:
 *	Nothing.
 */
void
irix_winfo(char *str)
{
	(void) strcat(str, "SGI Irix sound driver\n");
}

#endif	/* CDDA_WR_IRIX CDDA_SYSVIPC */

