/*
 *   cdda - CD Digital Audio support
 *
 *   Copyright (C) 1993-2002  Ti Kan
 *   E-mail: xmcd@amb.org
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *   Digital/Compaq/HP Tru64 UNIX (OSF1) MMS audio driver support
 */
#ifndef LINT
static char *_wr_osf1_c_ident_ = "@(#)wr_osf1.c	7.23 02/11/13";
#endif

#include "common_d/appenv.h"
#include "common_d/util.h"
#include "libdi_d/libdi.h"
#include "cdda_d/cdda.h"

#if defined(CDDA_WR_OSF1) && defined(CDDA_SYSVIPC)

#ifdef HAS_MMS
#include <mme/mme_api.h>
#endif
#include "cdda_d/sysvipc.h"
#include "cdda_d/sem.h"
#include "cdda_d/wr_gen.h"
#include "cdda_d/wr_osf1.h"


extern appdata_t	app_data;
extern FILE		*errfp;

#ifdef HAS_MMS
STATIC HWAVEOUT		osf1_devhandle = 0;	/* Audio device handle */
STATIC UINT		osf1_devid;		/* Audio device ID */
STATIC LPWAVEHDR	osf1_wavehdrp = NULL;	/* Wave header pointer */
#endif	/* HAS_MMS */

STATIC int		osf1_wsemid;		/* Semaphores identifier */
STATIC pid_t		osf1_pipe_pid = -1;	/* Pipe to program pid */
STATIC cd_state_t	*osf1_wcd;		/* Shared memory pointer */
STATIC bool_t		osf1_write_dev,		/* Write to audio device */
			osf1_write_file,	/* Write to output file */
			osf1_write_pipe,	/* Pipe to program */
			osf1_file_be = FALSE;	/* Big endian output file */
STATIC gen_desc_t	*osf1_file_desc = NULL,	/* Output file desc */
			*osf1_pipe_desc = NULL;	/* Pipe to program desc */
STATIC unsigned int	osf1_file_mode;		/* Output file mode */


#ifdef HAS_MMS
/*
 * osf1aud_wave_cb
 *	Callback function used to notify that playing is done on a
 *	block of audio data.
 *
 * Args:
 *	waveout - waveout handle
 *	msg - message code
 *	instance - instance number
 *	p1 - parameter 1
 *	p2 - parameter 2
 *
 * Return:
 *	Nothing.
 */
/*ARGSUSED*/
STATIC void
osf1aud_wave_cb(HANDLE waveout, UINT msg, DWORD instance, LPARAM p1, LPARAM p2)
{
	switch (msg) {
	case WOM_OPEN:
	case WOM_CLOSE:
	case WOM_DONE:
		break;
	default:
		DBGPRN(DBG_SND)(errfp,
			"osf1aud_wave_cb: unknown message %u received.\n",
			msg);
		break;
	}
}
#endif	/* HAS_MMS */


/*
 * osf1aud_open_dev
 *	Opens the audio device.
 *
 * Args:
 *	None.
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
STATIC bool_t
osf1aud_open_dev(void)
{
#ifndef HAS_MMS
	(void) sprintf(osf1_wcd->i->msgbuf,
		    "osf1aud_open_dev: MMS audio support not compiled-in");
	DBGPRN(DBG_SND)(errfp, "%s\n", osf1_wcd->i->msgbuf);
	return FALSE;
#else
	MMRESULT	status;
	LPWAVEOUTCAPS	caps;
	char		*cp;

	if (!gen_set_eid(osf1_wcd))
		return FALSE;

	if (waveOutGetNumDevs() < 1) {
		(void) sprintf(osf1_wcd->i->msgbuf,
			    "osf1aud_open_dev: audio disabled:\n%s",
			    "No MMS audio device available.");
		DBGPRN(DBG_SND)(errfp, "%s\n", osf1_wcd->i->msgbuf);
		(void) gen_reset_eid(osf1_wcd);
		return FALSE;
	}

	caps = (LPWAVEOUTCAPS) mmeAllocMem(sizeof(WAVEOUTCAPS));
	if (caps == NULL) {
		(void) sprintf(osf1_wcd->i->msgbuf,
			    "osf1aud_open_dev: Out of memory");
		DBGPRN(DBG_SND)(errfp, "%s\n", osf1_wcd->i->msgbuf);
		(void) gen_reset_eid(osf1_wcd);
		return FALSE;
	}

	/* Use AUDIODEV environment for device ID if specified by user,
	 * otherwise use 0.
	 */
	if ((cp = getenv("AUDIODEV")) != NULL)
		osf1_devid = (UINT) atoi(cp);
	else
		osf1_devid = 0;

	status = waveOutGetDevCaps(osf1_devid, caps, sizeof(WAVEOUTCAPS));
	if (status != MMSYSERR_NOERROR) {
		mmeFreeMem(caps);
		(void) sprintf(osf1_wcd->i->msgbuf,
			    "osf1aud_open_dev: "
			    "waveOutGetDevCaps failed (status=%d)", status);
		DBGPRN(DBG_SND)(errfp, "%s\n", osf1_wcd->i->msgbuf);
		(void) gen_reset_eid(osf1_wcd);
		return FALSE;
	}

	if ((caps->dwFormats & WAVE_FORMAT_4S16) == 0) {
		mmeFreeMem(caps);
		(void) sprintf(osf1_wcd->i->msgbuf,
			    "osf1aud_open_dev: The audio device does not\n"
			    "support 44.1 kHz Stereo 16-bit audio.");
		DBGPRN(DBG_SND)(errfp, "%s\n", osf1_wcd->i->msgbuf);
		(void) gen_reset_eid(osf1_wcd);
		return FALSE;
	}
	mmeFreeMem(caps);

	(void) gen_reset_eid(osf1_wcd);
	return TRUE;
#endif	/* HAS_MMS */
}


/*
 * osf1aud_close_dev
 *	Closes the audio device.
 *
 * Args:
 *	None.
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
STATIC bool_t
osf1aud_close_dev(void)
{
#ifndef HAS_MMS
	return FALSE;
#else
	MMRESULT	status;

	if (!gen_set_eid(osf1_wcd))
		return FALSE;

	if ((status = waveOutReset(osf1_devhandle)) != MMSYSERR_NOERROR) {
		DBGPRN(DBG_SND)(errfp, "osf1aud_close_dev: "
			    "waveOutReset failed (status=%d)\n", status);
	}
	if ((status = waveOutClose(osf1_devhandle)) != MMSYSERR_NOERROR) {
		DBGPRN(DBG_SND)(errfp, "osf1aud_close_dev: "
			    "waveOutClose failed (status=%d)\n", status);
	}

	DBGPRN(DBG_GEN|DBG_SND)(errfp, "\nClosed [audio:%u]\n", osf1_devid);

	osf1_devhandle = 0;
	mmeFreeMem(osf1_wavehdrp);
	osf1_wavehdrp = NULL;

	(void) gen_reset_eid(osf1_wcd);
	return ((bool_t) status == MMSYSERR_NOERROR);
#endif	/* HAS_MMS */
}


/*
 * osf1aud_config
 *	Sets up audio device for playing CD quality audio.
 *
 * Args:
 *	None.
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
STATIC bool_t
osf1aud_config(void)
{
#ifndef HAS_MMS
	(void) sprintf(osf1_wcd->i->msgbuf,
		    "osf1aud_config: MMS audio support not compiled-in");
	DBGPRN(DBG_SND)(errfp, "%s\n", osf1_wcd->i->msgbuf);
	return FALSE;
#else
	MMRESULT	status;
	LPPCMWAVEFORMAT	wp;

	wp = (LPPCMWAVEFORMAT) mmeAllocMem(sizeof(PCMWAVEFORMAT));
	if (wp == NULL) {
		(void) sprintf(osf1_wcd->i->msgbuf,
			    "osf1aud_config: Out of memory");
		DBGPRN(DBG_SND)(errfp, "%s\n", osf1_wcd->i->msgbuf);
		return FALSE;
	}

	wp->wf.nSamplesPerSec = 44100;
	wp->wf.nChannels = 2;
	wp->wBitsPerSample = 16;
	wp->wf.wFormatTag = WAVE_FORMAT_PCM;

	status = waveOutOpen(
		&osf1_devhandle,
		osf1_devid,
		(LPWAVEFORMAT) wp,
		osf1aud_wave_cb,
		NULL,
		WAVE_ALLOWSYNC | CALLBACK_FUNCTION
	);
	if (status != MMSYSERR_NOERROR) {
		(void) sprintf(osf1_wcd->i->msgbuf,
			    "osf1aud_config: waveOutOpen failed (status=%d)",
			    status);
		DBGPRN(DBG_SND)(errfp, "%s\n", osf1_wcd->i->msgbuf);
		return FALSE;
	}

	mmeFreeMem(wp);

	osf1_wavehdrp = (LPWAVEHDR) mmeAllocMem(sizeof(WAVEHDR));
	if (osf1_wavehdrp == NULL) {
		(void) waveOutClose(osf1_devhandle);
		osf1_devhandle = 0;
		(void) sprintf(osf1_wcd->i->msgbuf,
			    "osf1aud_config: Out of memory");
		DBGPRN(DBG_SND)(errfp, "%s\n", osf1_wcd->i->msgbuf);
		return FALSE;
	}

	osf1_wavehdrp->lpData = (LPSTR)
				mmeAllocMem(osf1_wcd->cds->chunk_bytes);
	if (osf1_wavehdrp->lpData == NULL) {
		(void) waveOutClose(osf1_devhandle);
		osf1_devhandle = 0;
		mmeFreeMem(osf1_wavehdrp);
		osf1_wavehdrp = NULL;
		(void) sprintf(osf1_wcd->i->msgbuf,
			    "osf1aud_config: Out of memory");
		DBGPRN(DBG_SND)(errfp, "%s\n", osf1_wcd->i->msgbuf);
		return FALSE;
	}

	DBGPRN(DBG_GEN|DBG_SND)(errfp, "\nOpened [audio:%u]\n", osf1_devid);

	return TRUE;
#endif	/* HAS_MMS */
}


/*
 * osf1aud_write_chunk
 *      Write a chunk of audio data to the audio device.  This function
 *      assumes 16-bit, 2-channel interleaved frames.
 *
 * Args:
 *      data - Pointer to the data buffer
 *      len - Number of bytes to write
 *
 * Return:
 *      FALSE - failure
 *      TRUE  - success
 */
STATIC bool_t
osf1aud_write_chunk(sword16_t *data, size_t len)
{
#ifndef HAS_MMS
	(void) sprintf(osf1_wcd->i->msgbuf,
		    "osf1aud_write_chunk: MMS audio support not compiled-in");
	DBGPRN(DBG_SND)(errfp, "%s\n", osf1_wcd->i->msgbuf);
	return FALSE;
#else
	MMRESULT	status;

	osf1_wavehdrp->dwBufferLength = len;
	(void) memcpy(osf1_wavehdrp->lpData, (byte_t *) data, len);

	DBGPRN(DBG_SND)(errfp, "\nWrite [audio:%u]: %d bytes\n",
			osf1_devid, len);

	status = waveOutWrite(osf1_devhandle, osf1_wavehdrp, sizeof(WAVEHDR));
	if (status != MMSYSERR_NOERROR) {
		(void) sprintf(osf1_wcd->i->msgbuf,
			    "osf1aud_write_chunk: "
			    "waveOutWrite failed (status=%d)", status);
		DBGPRN(DBG_SND)(errfp, "%s\n", osf1_wcd->i->msgbuf);
		return FALSE;
	}
	mmeWaitForCallbacks();
	mmeProcessCallbacks();

	return TRUE;
#endif	/* HAS_MMS */
}


/*
 * osf1aud_write
 *	Function responsible for writing from the buffer to the audio
 *	device (and file if saving).
 *
 * Args:
 *	s - Pointer to the curstat_t structure
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
STATIC bool_t
osf1aud_write(curstat_t *s)
{
	byte_t	*in_data,
		*sw_data,
		*le_data,
		*be_data,
		*dev_data,
		*file_data,
		*pipe_data;
	char	*fname = NULL;
	int	trk_idx = -1,
		hbcnt = 1;
	time_t	start_time,
		tot_time;
	bool_t	ret;

	/* Get memory for audio device write buffer */
	le_data = (byte_t *) MEM_ALLOC("le_data", osf1_wcd->cds->chunk_bytes);
	if (le_data == NULL) {
		(void) sprintf(osf1_wcd->i->msgbuf,
				"osf1aud_write: out of memory");
		DBGPRN(DBG_GEN)(errfp, "%s\n", osf1_wcd->i->msgbuf);
		return FALSE;
	}

	/* Get memory for audio file write buffer */
	be_data = (byte_t *) MEM_ALLOC("be_data", osf1_wcd->cds->chunk_bytes);
	if (be_data == NULL) {
		(void) sprintf(osf1_wcd->i->msgbuf,
				"osf1aud_write: out of memory");
		DBGPRN(DBG_GEN)(errfp, "%s\n", osf1_wcd->i->msgbuf);
		return FALSE;
	}

	(void) memset(le_data, 0, osf1_wcd->cds->chunk_bytes);
	(void) memset(be_data, 0, osf1_wcd->cds->chunk_bytes);

	in_data = app_data.cdda_bigendian ? be_data : le_data;
	sw_data = app_data.cdda_bigendian ? le_data : be_data;

	/* Start time */
	start_time = time(NULL);

	/* While not stopped */
	ret = TRUE;
	while (osf1_wcd->i->state != CDSTAT_COMPLETED) {
		/* Set the current track and track length info */
		cdda_wr_setcurtrk(osf1_wcd);

		/* Get lock */
		cdda_waitsem(osf1_wsemid, LOCK);

		/* End if finished and nothing in the buffer */
		if (osf1_wcd->cdb->occupied == 0 && osf1_wcd->i->cdda_done) {
			osf1_wcd->i->state = CDSTAT_COMPLETED;
			cdda_postsem(osf1_wsemid, LOCK);
			break;
		}

		/* Wait until there is something there */
		while (osf1_wcd->cdb->occupied <= 0 &&
		       osf1_wcd->i->state != CDSTAT_COMPLETED) {
			cdda_postsem(osf1_wsemid, LOCK);
			cdda_waitsem(osf1_wsemid, DATA);
			cdda_waitsem(osf1_wsemid, LOCK);
		}

		/* Break if stopped */
		if (osf1_wcd->i->state == CDSTAT_COMPLETED) {
			cdda_postsem(osf1_wsemid, LOCK);
			break;
		}

		/* Copy a chunk from the nextout slot into the data buffers */
		(void) memcpy(
			in_data,
			&osf1_wcd->cdb->b[
			    osf1_wcd->cds->chunk_bytes * osf1_wcd->cdb->nextout
			],
			osf1_wcd->cds->chunk_bytes
		);

		/* Update pointers */
		osf1_wcd->cdb->nextout++;
		osf1_wcd->cdb->nextout %= osf1_wcd->cds->buffer_chunks;
		osf1_wcd->cdb->occupied--;

		/* Set heartbeat and interval, update stats */
		if (--hbcnt == 0) {
			cdda_heartbeat(&osf1_wcd->i->writer_hb,CDDA_HB_WRITER);

			tot_time = time(NULL) - start_time;
			if (tot_time > 0 && osf1_wcd->i->frm_played > 0) {
				osf1_wcd->i->frm_per_sec = (int)
					((float) osf1_wcd->i->frm_played /
					 tot_time) + 0.5;
			}

			hbcnt = cdda_heartbeat_interval(
				osf1_wcd->i->frm_per_sec
			);
		}

		/* Signal room available */
		cdda_postsem(osf1_wsemid, ROOM);

		/* Release lock */
		cdda_postsem(osf1_wsemid, LOCK);

		/* Change channel routing if necessary */
		gen_chroute(osf1_wcd->i->chroute, (sword16_t *)(void *) in_data,
			    (size_t) osf1_wcd->cds->chunk_bytes);

		/* Prepare byte-swapped version of the data */
		gen_byteswap(in_data, sw_data,
			     (size_t) osf1_wcd->cds->chunk_bytes);

		/* Write to device */
		if (osf1_write_dev) {
			/* Always send little endian data */
			dev_data = le_data;

			if (!osf1aud_write_chunk((sword16_t *) dev_data,
					osf1_wcd->cds->chunk_bytes)) {
				osf1_wcd->i->state = CDSTAT_COMPLETED;
				cdda_postsem(osf1_wsemid, LOCK);
				ret = FALSE;
				break;
			}
		}

		/* Write to file */
		if (osf1_write_file) {
			file_data = osf1_file_be ? be_data : le_data;

			/* Open new track file if necessary */
			if (!app_data.cdda_trkfile) {
				fname = s->trkinfo[0].outfile;
			}
			else if (trk_idx != osf1_wcd->i->trk_idx) {
				trk_idx = osf1_wcd->i->trk_idx;

				if (osf1_file_desc != NULL &&
				    !gen_close_file(osf1_file_desc)) {
					osf1_wcd->i->state = CDSTAT_COMPLETED;
					cdda_postsem(osf1_wsemid, LOCK);
					ret = FALSE;
					break;
				}

				fname = s->trkinfo[trk_idx].outfile;
				osf1_file_desc = gen_open_file(
					osf1_wcd,
					fname,
					O_WRONLY | O_TRUNC | O_CREAT,
					(mode_t) osf1_file_mode,
					app_data.cdda_filefmt,
					osf1_wcd->i->trk_len
				);
				if (osf1_file_desc == NULL) {
					osf1_wcd->i->state = CDSTAT_COMPLETED;
					cdda_postsem(osf1_wsemid, LOCK);
					ret = FALSE;
					break;
				}
			}

			if (!gen_write_chunk(osf1_file_desc, file_data,
					(size_t) osf1_wcd->cds->chunk_bytes)) {
				osf1_wcd->i->state = CDSTAT_COMPLETED;
				cdda_postsem(osf1_wsemid, LOCK);
				ret = FALSE;
				break;
			}
		}

		/* Pipe to program */
		if (osf1_write_pipe) {
			pipe_data = osf1_file_be ? be_data : le_data;

			if (!gen_write_chunk(osf1_pipe_desc, pipe_data,
					(size_t) osf1_wcd->cds->chunk_bytes)) {
				osf1_wcd->i->state = CDSTAT_COMPLETED;
				cdda_postsem(osf1_wsemid, LOCK);
				ret = FALSE;
				break;
			}
		}

		osf1_wcd->i->frm_played += osf1_wcd->cds->chunk_blocks;

		while (osf1_wcd->i->state == CDSTAT_PAUSED)
			util_delayms(400);

		/* Update debug level */
		app_data.debug = osf1_wcd->i->debug;
	}

	/* Wake up reader */
	cdda_postsem(osf1_wsemid, ROOM);

	/* Free memory */
	MEM_FREE(le_data);
	MEM_FREE(be_data);

	return (ret);
}


/*
 * osf1aud_cleanup
 *	Detaches shared memory.
 *
 * Args:
 *	killreader - Whether to kill the reader process
 *
 * Return:
 *	Nothing.
 */
STATIC void
osf1aud_cleanup(bool_t killreader)
{
	DBGPRN(DBG_GEN)(errfp,
			  "\nosf1aud_cleanup: Cleaning up writer pid=%d\n",
			  (int) getpid());

#ifdef HAS_MMS
	if (osf1_wavehdrp != NULL)
		(void) osf1aud_close_dev();
#endif

	if (osf1_file_desc != NULL) {
		(void) gen_close_file(osf1_file_desc);
		osf1_file_desc = NULL;
	}

	if (osf1_pipe_desc != NULL) {
		(void) gen_close_pipe(osf1_pipe_desc);
		osf1_pipe_desc = NULL;
	}

	(void) sleep(1);

	if (osf1_pipe_pid > 0) {
		(void) kill(osf1_pipe_pid, SIGTERM);
		osf1_pipe_pid = -1;
	}

	if (osf1_wcd != NULL) {
		if (osf1_wcd->i != NULL) {
			if (killreader && osf1_wcd->i->reader != 0) {
				(void) kill(osf1_wcd->i->reader, SIGTERM);
				osf1_wcd->i->state = CDSTAT_COMPLETED;
			}

			/* Reset frames played counter */
			osf1_wcd->i->frm_played = 0;
		}

		MEM_FREE(osf1_wcd);
		osf1_wcd = NULL;
	}
}


/*
 * osf1aud_onterm
 *	SIGTERM and SIGPIPE signal handler - gracefully exit
 *
 * Args:
 *	signo - The signal number
 *
 * Return:
 *	Nothing.
 */
STATIC void
osf1aud_onterm(int signo)
{
	osf1aud_cleanup((bool_t) (signo == SIGPIPE));
	exit(1);
}


/*
 * osf1_writeinit
 *	Pre-playback support check function
 *
 * Args:
 *	None.
 *
 * Return:
 *	Bitmask of supported features
 */
word32_t
osf1_writeinit(void)
{
#ifdef HAS_MMS
	return (CDDA_WRITEDEV | CDDA_WRITEFILE | CDDA_WRITEPIPE);
#else
	return (CDDA_WRITEFILE | CDDA_WRITEPIPE);
#endif
}


/*
 * osf1_write
 *	Opens audio device/file, attaches shared memory and semaphores.
 *	Continuously reads data from shared memory and writes it
 *	to the audio device/file.
 *
 * Args:
 *	s    - Pointer to the curstat_t structure
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
bool_t
osf1_write(curstat_t *s)
{
	size_t	estlen;

	if (!PLAYMODE_IS_CDDA(app_data.play_mode)) {
		(void) fprintf(errfp, "osf1_write: Nothing to do.\n");
		return FALSE;
	}

#ifdef HAS_MMS
	osf1_write_dev  = (bool_t) ((app_data.play_mode & PLAYMODE_CDDA) != 0);
#else
	osf1_write_dev  = FALSE;
#endif
	osf1_write_file = (bool_t) ((app_data.play_mode & PLAYMODE_FILE) != 0);
	osf1_write_pipe = (bool_t) ((app_data.play_mode & PLAYMODE_PIPE) != 0);

	/* Generic services initialization */
	gen_write_init();

	/* Set some signal behaviors */
	(void) signal(SIGINT, SIG_IGN);
	(void) signal(SIGQUIT, SIG_IGN);
	(void) signal(SIGALRM, SIG_IGN);
	(void) signal(SIGUSR1, SIG_IGN);
	(void) signal(SIGPIPE, osf1aud_onterm);
	(void) signal(SIGTERM, osf1aud_onterm);

	/* Allocate memory */
	osf1_wcd = (cd_state_t *) MEM_ALLOC("cd_state_t", sizeof(cd_state_t));
	if (osf1_wcd == NULL) {
		(void) fprintf(errfp, "osf1_write: out of memory\n");
		return FALSE;
	}

	(void) memset(osf1_wcd, 0, sizeof(cd_state_t));

	/* Initialize shared memory and semaphores */
	if ((osf1_wsemid = cdda_rw_initipc(osf1_wcd)) < 0) {
		osf1aud_cleanup(TRUE);
		return FALSE;
	}

	(void) sscanf(app_data.cdinfo_filemode, "%o", &osf1_file_mode);
	/* Make sure file is at least accessible by user */
	osf1_file_mode |= S_IRUSR | S_IWUSR;
	osf1_file_mode &= ~(S_ISUID | S_ISGID | S_IWGRP | S_IWOTH);

	estlen = (osf1_wcd->i->end_lba - osf1_wcd->i->start_lba + 1) *
		 CDDA_BLKSZ;

	/* Set heartbeat */
	cdda_heartbeat(&osf1_wcd->i->writer_hb, CDDA_HB_WRITER);

	if (osf1_write_file || osf1_write_pipe) {
		/* Open file and/or pipe */
		if (!app_data.cdda_trkfile && osf1_write_file) {
			osf1_file_desc = gen_open_file(
				osf1_wcd,
				s->trkinfo[0].outfile,
				O_WRONLY | O_TRUNC | O_CREAT,
				(mode_t) osf1_file_mode,
				app_data.cdda_filefmt,
				estlen
			);
			if (osf1_file_desc == NULL) {
				osf1aud_cleanup(TRUE);
				return FALSE;
			}
		}

		if (osf1_write_pipe) {
			osf1_pipe_desc = gen_open_pipe(
				osf1_wcd,
				app_data.pipeprog,
				&osf1_pipe_pid,
				app_data.cdda_filefmt,
				estlen
			);
			if (osf1_pipe_desc == NULL) {
				osf1aud_cleanup(TRUE);
				return FALSE;
			}
			DBGPRN(DBG_SND)(errfp,
					  "\nosf1aud_write: Pipe to [%s]: "
					  "chunk_blks=%d\n",
					  app_data.pipeprog,
					  osf1_wcd->cds->chunk_blocks);
		}

		switch (app_data.cdda_filefmt) {
		case FILEFMT_AU:
		case FILEFMT_AIFF:
		case FILEFMT_AIFC:
			/* Always big endian */
			osf1_file_be = TRUE;
			break;
		case FILEFMT_WAV:
		case FILEFMT_OGG:
			/* Always little endian */
			osf1_file_be = FALSE;
			break;
		case FILEFMT_MP3:
#ifdef USE_LIBMP3LAME
			/* Use native endian */
#if _BYTE_ORDER_ == _B_ENDIAN_
			osf1_file_be = TRUE;
#else
			osf1_file_be = FALSE;
#endif
#else
			/* Use big endian */
			osf1_file_be = TRUE;
#endif
			break;
		default:
			break;
		}
	}

#ifdef HAS_MMS
	if (osf1_write_dev) {
		/* Open audio device */
		if (!osf1aud_open_dev()) {
			osf1aud_cleanup(TRUE);
			return FALSE;
		}

		/* Configure audio */
		if (!osf1aud_config()) {
			osf1aud_cleanup(TRUE);
			return FALSE;
		}
	}
#endif

	/* Do the writing */
	if (!osf1aud_write(s)) {
		osf1aud_cleanup(TRUE);
		return FALSE;
	}

	if (osf1_write_file && !gen_close_file(osf1_file_desc)) {
		osf1aud_cleanup(TRUE);
		return FALSE;
	}
	osf1_file_desc = NULL;

	if (osf1_write_pipe && !gen_close_pipe(osf1_pipe_desc)) {
		osf1aud_cleanup(TRUE);
		return FALSE;
	}
	osf1_pipe_desc = NULL;
	osf1_pipe_pid = -1;

	osf1aud_cleanup(FALSE);
	return TRUE;
}


/*
 * osf1_winfo
 *	Append method identification to informational string.
 *
 * Args:
 *	str - The informational string
 *
 * Return:
 *	Nothing.
 */
void
osf1_winfo(char *str)
{
	(void) strcat(str, "Tru64 UNIX (OSF1) MMS audio driver\n");
}

#endif	/* CDDA_WR_OSF1 CDDA_SYSVIPC */

